@@ -1,12 +1,11 @@
 /*  main.c
  *  Created by Mengyao Zhao on 06/23/11.
- *	Version 1.0
- *  Last revision by Mengyao Zhao on 07/19/16.
+ *	Version 1.2.2
+ *  Last revision by Mengyao Zhao on 2022-May-21.
  */
 
 #include <stdlib.h>
 #include <stdint.h>
-#include <emmintrin.h>
 #include <zlib.h>
 #include <stdio.h>
 #include <time.h>
@@ -16,6 +15,13 @@
 #include "ssw.h"
 #include "kseq.h"
 
+#ifdef __ARM_NEON // (M1)
+#include "sse2neon.h"
+#else // x86 (Intel)
+#include <emmintrin.h>
+#endif
+
+
 #ifdef __GNUC__
 #define LIKELY(x) __builtin_expect((x),1)
 #define UNLIKELY(x) __builtin_expect((x),0)
@@ -36,14 +42,14 @@ KSEQ_INIT(gzFile, gzread)
 static void reverse_comple(const char* seq, char* rc) {
 	int32_t end = strlen(seq), start = 0;
 	static const int8_t rc_table[128] = {
-		4, 4,  4, 4,  4,  4,  4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
-		4, 4,  4, 4,  4,  4,  4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
-		4, 4,  4, 4,  4,  4,  4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
-		4, 4,  4, 4,  4,  4,  4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
-		4, 84, 4, 71, 4,  4,  4, 67, 4, 4, 4, 4,  4, 4, 4, 4,
-		4, 4,  4, 4,  65, 65, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
-		4, 84, 4, 71, 4,  4,  4, 67, 4, 4, 4, 4,  4, 4, 4, 4,
-		4, 4,  4, 4,  65, 65, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4
+		4, 4,  4, 4,  4,  4,  4, 4,  4, 4, 4, 4,  4, 4, 4,  4,
+		4, 4,  4, 4,  4,  4,  4, 4,  4, 4, 4, 4,  4, 4, 4,  4,
+		4, 4,  4, 4,  4,  4,  4, 4,  4, 4, 4, 4,  4, 4, 4,  4,
+		4, 4,  4, 4,  4,  4,  4, 4,  4, 4, 4, 4,  4, 4, 4,  4,
+		4, 84, 4, 71, 4,  4,  4, 67, 4, 4, 4, 4,  4, 4, 78, 4,
+		4, 4,  4, 4,  65, 65, 4, 4,  4, 4, 4, 4,  4, 4, 4,  4,
+		4, 84, 4, 71, 4,  4,  4, 67, 4, 4, 4, 4,  4, 4, 78,  4,
+		4, 4,  4, 4,  65, 65, 4, 4,  4, 4, 4, 4,  4, 4, 4,  4
 	};
 	rc[end] = '\0';
 	-- end;
@@ -60,6 +66,8 @@ static void ssw_write (s_align* a,
 			const kseq_t* ref_seq,
 			const kseq_t* read,
 			const char* read_seq,	// strand == 0: original read; strand == 1: reverse complement read
+			const int8_t* ref_num,
+			const int8_t* read_num,
 			const int8_t* table,
 			int8_t strand,	// 0: forward aligned ; 1: reverse complement aligned
 			int8_t sam) {	// 0: Blast like output; 1: Sam format output
@@ -155,56 +163,28 @@ end:
 		fprintf(stdout, "%s\t", read->name.s);
 		if (a->score1 == 0) fprintf(stdout, "4\t*\t0\t255\t*\t*\t0\t0\t*\t*\n");
 		else {
-			int32_t c, l = a->read_end1 - a->read_begin1 + 1, qb = a->ref_begin1, pb = a->read_begin1, p;
+			int32_t c, p;
 			uint32_t mapq = -4.343 * log(1 - (double)abs(a->score1 - a->score2)/(double)a->score1);
 			mapq = (uint32_t) (mapq + 4.99);
 			mapq = mapq < 254 ? mapq : 254;
 			if (strand) fprintf(stdout, "16\t");
 			else fprintf(stdout, "0\t");
 			fprintf(stdout, "%s\t%d\t%d\t", ref_seq->name.s, a->ref_begin1 + 1, mapq);
-			mismatch = mark_mismatch(a->ref_begin1, a->read_begin1, a->read_end1, ref_seq->seq.s, read_seq, read->seq.l, &a->cigar, &a->cigarLen);
+			mismatch = mark_mismatch(a->ref_begin1, a->read_begin1, a->read_end1, ref_num, read_num, read->seq.l, &a->cigar, &a->cigarLen);
 			for (c = 0; c < a->cigarLen; ++c) {
 				char letter = cigar_int_to_op(a->cigar[c]);
 				uint32_t length = cigar_int_to_len(a->cigar[c]);
 				fprintf(stdout, "%lu%c", (unsigned long)length, letter);
 			}
-	//		fprintf(stderr, "%s\tmismatch: %d\n", read->name.s, mismatch);
 			fprintf(stdout, "\t*\t0\t0\t");
-			for (c = a->read_begin1; c <= a->read_end1; ++c) fprintf(stdout, "%c", read_seq[c]);
+			fprintf(stdout, "%s", read_seq);
 			fprintf(stdout, "\t");
 			if (read->qual.s && strand) {
-				p = a->read_end1;
-				for (c = 0; c < l; ++c) {
-					fprintf(stdout, "%c", read->qual.s[p]);
-					--p;
-				}
-			}else if (read->qual.s){
-				p = a->read_begin1;
-				for (c = 0; c < l; ++c) {
-					fprintf(stdout, "%c", read->qual.s[p]);
-					++p;
-				}
-			} else fprintf(stdout, "*");
+				for (p = read->qual.l - 1; p >= 0; --p) fprintf(stdout, "%c", read->qual.s[p]);
+			}else if (read->qual.s) fprintf (stdout, "%s", read->qual.s);
+			else fprintf(stdout, "*");
 			fprintf(stdout, "\tAS:i:%d", a->score1);
-			mapq = 0;	// counter of difference
-			for (c = 0; c < a->cigarLen; ++c) {
-				char letter = cigar_int_to_op(a->cigar[c]);
-				uint32_t length = cigar_int_to_len(a->cigar[c]);
-				if (letter == 'M') {
-					for (p = 0; p < length; ++p){
-						if (table[(int)*(ref_seq->seq.s + qb)] != table[(int)*(read_seq + pb)]) ++mapq;
-						++qb;
-						++pb;
-					}
-				} else if (letter == 'I') {
-					pb += length;
-					mapq += length;
-				} else {
-					qb += length;
-					mapq += length;
-				}
-			}
-			fprintf(stdout,"\tNM:i:%d\t", mapq);
+			fprintf(stdout,"\tNM:i:%d\t", mismatch);
 			if (a->score2 > 0) fprintf(stdout, "ZS:i:%d\n", a->score2);
 			else fprintf(stdout, "\n");
 		}
@@ -217,13 +197,9 @@ int main (int argc, char * const argv[]) {
 	gzFile read_fp, ref_fp;
 	kseq_t *read_seq, *ref_seq;
 	int32_t l, m, k, match = 2, mismatch = 2, gap_open = 3, gap_extension = 1, path = 0, reverse = 0, n = 5, sam = 0, protein = 0, header = 0, s1 = 67108864, s2 = 128, filter = 0;
-	int8_t* mata = (int8_t*)calloc(25, sizeof(int8_t));
-	const int8_t* mat = mata;
-	char mat_name[16];
-	mat_name[0] = '\0';
-	int8_t* ref_num = (int8_t*)malloc(s1);
-	int8_t* num = (int8_t*)malloc(s2), *num_rc = 0;
-	char* read_rc = 0;
+    int8_t *mata, *ref_num, *num, *num_rc = 0;
+    const int8_t* mat;
+	char* read_rc = 0, *mat_name = 0;
 
 	static const int8_t mat50[] = {
 	//  A   R   N   D   C   Q   E   G   H   I   L   K   M   F   P   S   T   W   Y   V   B   Z   X   *
@@ -279,22 +255,6 @@ int main (int argc, char * const argv[]) {
 
 	int8_t* table = nt_table;
 
-	// Parse command line.
-	while ((l = getopt(argc, argv, "m:x:o:e:a:f:pcrsh")) >= 0) {
-		switch (l) {
-			case 'm': match = atoi(optarg); break;
-			case 'x': mismatch = atoi(optarg); break;
-			case 'o': gap_open = atoi(optarg); break;
-			case 'e': gap_extension = atoi(optarg); break;
-			case 'a': strcpy(mat_name, optarg); break;
-			case 'f': filter = atoi(optarg); break;
-			case 'p': protein = 1; break;
-			case 'c': path = 1; break;
-			case 'r': reverse = 1; break;
-			case 's': sam = 1; break;
-			case 'h': header = 1; break;
-		}
-	}
 	if (optind + 2 > argc) {
 		fprintf(stderr, "\n");
 		fprintf(stderr, "Usage: ssw_test [options] ... <target.fasta> <query.fasta>(or <query.fastq>)\n");
@@ -313,21 +273,52 @@ int main (int argc, char * const argv[]) {
 		return 1;
 	}
 
+	// Parse command line.
+	while ((l = getopt(argc, argv, "m:x:o:e:a:f:pcrsh")) >= 0) {
+		switch (l) {
+			case 'm': match = atoi(optarg); break;
+			case 'x': mismatch = atoi(optarg); break;
+			case 'o': gap_open = atoi(optarg); break;
+			case 'e': gap_extension = atoi(optarg); break;
+
+			case 'a': 
+                mat_name = (char*)malloc(strlen(optarg) + 1);
+                strcpy(mat_name, optarg); 
+                break;
+
+			case 'f': filter = atoi(optarg); break;
+			case 'p': protein = 1; break;
+			case 'c': path = 1; break;
+			case 'r': reverse = 1; break;
+			case 's': sam = 1; break;
+			case 'h': header = 1; break;
+		}
+	}
+
 	// initialize scoring matrix for genome sequences
+    mata = (int8_t*)calloc(25, sizeof(int8_t));
 	for (l = k = 0; LIKELY(l < 4); ++l) {
 		for (m = 0; LIKELY(m < 4); ++m) mata[k++] = l == m ? match : -mismatch;	/* weight_match : -weight_mismatch */
 		mata[k++] = 0; // ambiguous base
 	}
 	for (m = 0; LIKELY(m < 5); ++m) mata[k++] = 0;
+    mat = mata;
 
-	if (protein == 1 && (! strcmp(mat_name, "\0"))) {
+	if (protein == 1 && mat_name == 0) {
 		n = 24;
 		table = aa_table;
 		mat = mat50;
-	} else if (strcmp(mat_name, "\0")) {
+	} else if (mat_name != 0) {
 
 	// Parse score matrix.
 		FILE *f_mat = fopen(mat_name, "r");
+        free(mat_name);
+        if (f_mat == NULL) { 
+            fprintf(stderr, "Failed to open the weight matrix file.\n");
+            free(mata);
+            return 1;
+        }
+
 		char line[128];
 		mata = (int8_t*)realloc(mata, 1024 * sizeof(int8_t));
 		k = 0;
@@ -359,6 +350,7 @@ int main (int argc, char * const argv[]) {
 		}
 		if (k == 0) {
 			fprintf(stderr, "Problem of reading the weight matrix file.\n");
+            free(mata);
 			return 1;
 		}
 		fclose(f_mat);
@@ -367,7 +359,6 @@ int main (int argc, char * const argv[]) {
 		mat = mata;
 	}
 
-	//fprintf(stderr, "query: %s\n", argv[optind + 1]);
 	read_fp = gzopen(argv[optind + 1], "r");
 
     if (! read_fp) {
@@ -389,6 +380,8 @@ int main (int argc, char * const argv[]) {
 	}
 
 	// alignment
+	ref_num = (int8_t*)malloc(s1);
+	num = (int8_t*)malloc(s2);
 	if (reverse == 1 && n == 5) {
 		read_rc = (char*)malloc(s2);
 		num_rc = (int8_t*)malloc(s2);
@@ -416,7 +409,13 @@ int main (int argc, char * const argv[]) {
 			p_rc = ssw_init(num_rc, readLen, mat, n, 2);
 		}else if (reverse == 1 && n == 24) {
 			fprintf (stderr, "Reverse complement alignment is not available for protein sequences. \n");
-			return 1;
+            free(num);
+            free(ref_num);
+            if (num_rc) {
+                free(num_rc);
+                free(read_rc);
+            }
+            return 1;
 		}
 
 		ref_fp = gzopen(argv[optind], "r");
@@ -436,12 +435,17 @@ int main (int argc, char * const argv[]) {
 			if (reverse == 1 && protein == 0)
 				result_rc = ssw_align(p_rc, ref_num, refLen, gap_open, gap_extension, flag, filter, 0, maskLen);
 			if (result_rc && result_rc->score1 > result->score1 && result_rc->score1 >= filter) {
-				if (sam) ssw_write (result_rc, ref_seq, read_seq, read_rc, table, 1, 1);
-				else ssw_write (result_rc, ref_seq, read_seq, read_rc, table, 1, 0);
+                if (result_rc->flag == 2) fprintf(stderr, "Warning: The reverse compliment alignment of the following sequences may miss a small part.\nref_seq: %s\nread_seq: %s\n\n", ref_seq->name.s, read_seq->name.s);
+				if (sam) ssw_write (result_rc, ref_seq, read_seq, read_rc, ref_num, num_rc, table, 1, 1);
+				else ssw_write (result_rc, ref_seq, read_seq, read_rc, ref_num, num_rc, table, 1, 0);
 			}else if (result && result->score1 >= filter){
-				if (sam) ssw_write(result, ref_seq, read_seq, read_seq->seq.s, table, 0, 1);
-				else ssw_write(result, ref_seq, read_seq, read_seq->seq.s, table, 0, 0);
-			} else if (! result) return 1;
+                if (result->flag == 2) fprintf(stderr, "Warning: The alignment of the following sequences may miss a small part.\nref_seq: %s\nread_seq: %s\n\n", ref_seq->name.s, read_seq->name.s);
+				if (sam) ssw_write(result, ref_seq, read_seq, read_seq->seq.s, ref_num, num, table, 0, 1);
+				else ssw_write(result, ref_seq, read_seq, read_seq->seq.s, ref_num, num, table, 0, 0);
+			} else if (! result) {
+                fprintf(stderr, "Warning: Alignment between the following sequences is failed.\nref_seq: %s\nread_seq: %s\n\n", ref_seq->name.s, read_seq->name.s);
+                continue;
+            }
 			if (result_rc) align_destroy(result_rc);
 			align_destroy(result);
 		}
