@@ -1,3 +1,7 @@
+// ssw_cpp.cpp
+// Created by Wan-Ping Lee
+// Last revision by Mengyao Zhao on 2023-Apr-21
+
 #include "ssw_cpp.h"
 #include "ssw.h"
 
@@ -286,25 +290,18 @@ int Aligner::SetReferenceSequence(const char* seq, const int& length) {
 
   int len = 0;
   if (translation_matrix_) {
-    // calculate the valid length
-    //int calculated_ref_length = static_cast<int>(strlen(seq));
-    //int valid_length = (calculated_ref_length > length)
-    //                   ? length : calculated_ref_length;
-    int valid_length = length;
     // delete the current buffer
     CleanReferenceSequence();
     // allocate a new buffer
-    translated_reference_ = new int8_t[valid_length];
+    translated_reference_ = new int8_t[length];
 
-    len = TranslateBase(seq, valid_length, translated_reference_);
+    len = TranslateBase(seq, length, translated_reference_);
   } else {
     // nothing
   }
 
   reference_length_ = len;
   return len;
-
-
 }
 
 int Aligner::TranslateBase(const char* bases, const int& length,
@@ -322,8 +319,8 @@ int Aligner::TranslateBase(const char* bases, const int& length,
 }
 
 
-bool Aligner::Align(const char* query, const Filter& filter,
-                    Alignment* alignment) const
+uint16_t Aligner::Align(const char* query, const Filter& filter,
+                    Alignment* alignment, const int32_t maskLen) const
 {
   if (!translation_matrix_) return false;
   if (reference_length_ == 0) return false;
@@ -342,24 +339,24 @@ bool Aligner::Align(const char* query, const Filter& filter,
   s_align* s_al = ssw_align(profile, translated_reference_, reference_length_,
                                  static_cast<int>(gap_opening_penalty_),
 				 static_cast<int>(gap_extending_penalty_),
-				 flag, filter.score_filter, filter.distance_filter, query_len);
+				 flag, filter.score_filter, filter.distance_filter, maskLen);
 
   alignment->Clear();
   ConvertAlignment(*s_al, query_len, alignment);
   alignment->mismatches = CalculateNumberMismatch(&*alignment, translated_reference_, translated_query, query_len);
-
+  uint16_t align_flag = s_al->flag;
 
   // Free memory
   delete [] translated_query;
   align_destroy(s_al);
   init_destroy(profile);
 
-  return true;
+  return align_flag;
 }
 
 
-bool Aligner::Align(const char* query, const char* ref, const int& ref_len,
-                    const Filter& filter, Alignment* alignment) const
+uint16_t Aligner::Align(const char* query, const char* ref, const int& ref_len,
+                    const Filter& filter, Alignment* alignment, const int32_t maskLen) const
 {
   if (!translation_matrix_) return false;
 
@@ -369,9 +366,6 @@ bool Aligner::Align(const char* query, const char* ref, const int& ref_len,
   TranslateBase(query, query_len, translated_query);
 
   // calculate the valid length
-  //int calculated_ref_length = static_cast<int>(strlen(ref));
-  //int valid_ref_len = (calculated_ref_length > ref_len)
-  //                    ? ref_len : calculated_ref_length;
   int valid_ref_len = ref_len;
   int8_t* translated_ref = new int8_t[valid_ref_len];
   TranslateBase(ref, valid_ref_len, translated_ref);
@@ -386,11 +380,12 @@ bool Aligner::Align(const char* query, const char* ref, const int& ref_len,
   s_align* s_al = ssw_align(profile, translated_ref, valid_ref_len,
                                  static_cast<int>(gap_opening_penalty_),
 				 static_cast<int>(gap_extending_penalty_),
-				 flag, filter.score_filter, filter.distance_filter, query_len);
+				 flag, filter.score_filter, filter.distance_filter, maskLen);
 
   alignment->Clear();
   ConvertAlignment(*s_al, query_len, alignment);
   alignment->mismatches = CalculateNumberMismatch(&*alignment, translated_ref, translated_query, query_len);
+  uint16_t align_flag = s_al->flag;
 
   // Free memory
   delete [] translated_query;
@@ -398,7 +393,7 @@ bool Aligner::Align(const char* query, const char* ref, const int& ref_len,
   align_destroy(s_al);
   init_destroy(profile);
 
-  return true;
+  return align_flag;
 }
 
 void Aligner::Clear(void) {
